%{
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#define KITE_COMPILER
#include "objs/kite_object.h"
#include "kite_vm.h"
#include "../kite-vm-1.0/kite_opcodes.h"

#define YYPARSE_PARAM parm
#define YYLEX_PARAM parm

/*
 * Dirty, dirty hack to make yyerror() work with Kite's exception system.
 * Credit: http://www.everything2.com/index.pl?node_id=1509675
 */
#define yyerror( msg )  dirty_yyerror( parm, msg )
void dirty_yyerror(void *, char *);
void checkReserved(void *, char *);

/* 
 * For some weird reason, we had to split the macro out this way
 * in order for the file/line numbering in exceptions to work properly.
 */
extern void *yyget_extra(void *);
extern int yylex(void *, void *, void *);

/* Need to grab dylib path from compiler. TODO: use separate method to manipulate. */
extern char *kite_dylib_path;
%}

%token IMPORT
%token METHOD
%token SYMBOL_NAME
%token FROM
%token CLASS
%token CONSTRUCT
%token DESTRUCT
%token PROPERTY
%token GLOBAL
%token PLUS
%token MINUS
%token MULTIPLY
%token DIVIDE
%token UNARY_PLUS
%token UNARY_MINUS
%token MAP
%token REDUCE
%token ARRAY_DEREF
%token EQUALS
%token NOT_EQUALS
%token LESS_THAN
%token GREATER_THAN
%token LESS_OR_EQUALS
%token GREATER_OR_EQUALS
%token AND
%token OR
%token NOT
%token STRING_VALUE
%token FLOAT_VALUE
%token INT_VALUE
%token MOD
%token OPERATOR
%token XOR
%token LEFT_SHIFT
%token RIGHT_SHIFT
%token EVAL
%token DECIDE
%token BOOL_VALUE
%token UNTIL
%token WHILE
%token RUN
%token CATCH
%token IS
%token ISOF
%token MAKE
%token METHOD_CALL
%token NULL_VALUE
%token DOCSTRING
%token BREAK
%token CONTINUE
%token RETURN
%token VERSION_KEYWORD
%token REGEX_VALUE
%token REGEX_VALUE_CI
%token ARRAY_SET
%token OPTIONS_SET

/* Needed since we can't quite lay out the associativity using the rules yet. */
%left MAP
%left REDUCE

%error-verbose
%defines
%pure-parser
%locations
%union {
    double doubleValue;
    long intValue;
    char *stringValue;
    struct kite_object_t *internalValue;
    void *opValue;
}

%%

input:    /* empty */
    | DOCSTRING {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_object_t *obj = kite_top_stack(compiler->obj_created);
        kite_set_docstring(obj, $<stringValue>1);
        free($<stringValue>1);
    } input 
    | VERSION_KEYWORD STRING_VALUE ';' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        char *tmp;
        
        if (compiler->verMajor)
        {
            yyerror("Cannot specify required version more than once.");
            YYERROR;
        }
        
        compiler->verMajor = atoi($<stringValue>2);
        tmp = strstr($<stringValue>2, ".");
        if (!tmp)
        {
            yyerror("Parse error while reading version string");
            YYERROR;
        }
        compiler->verPatch = atoi(tmp);
        tmp = strstr(tmp, ".");
        if (!tmp)
        {
            compiler->verPatch = 0;
        }
        
        if (compiler->verMajor > VERSION_MAJOR ||
           (compiler->verMajor == VERSION_MAJOR && compiler->verMinor > VERSION_MINOR) ||
           (compiler->verMajor == VERSION_MAJOR && compiler->verMinor == VERSION_MINOR && compiler->verPatch > VERSION_PATCH))
        {
            char buf[1024];
#ifndef WIN32
            snprintf(buf, 1024, "Module can only run in Kite version %d.%d.%d and above.",
                    compiler->verMajor, compiler->verMinor, compiler->verPatch);
#else
            sprintf_s(buf, 1024, "Module can only run in Kite version %d.%d.%d and above.",
                    compiler->verMajor, compiler->verMinor, compiler->verPatch);
#endif /* WIN32 */
            yyerror(buf);
            YYERROR;    
        }
    } input
    | statement ';' input
    ;

statement:
    | data_manipulation_statement
    | options_set_statement
    | loop_statement
    | from_statement
    | import_statement
    | break_statement
    | continue_statement
    | return_statement
    | class_definition
    | property_definition
    | decide_statement
    | run_catch_statement
    | constructor_definition
    | destructor_definition
    | operator_definition
    ;

options_set_statement:  OPTIONS_SET STRING_VALUE STRING_VALUE {
    if (strcmp("import_path", $<stringValue>2) == 0)
    {
        /* TODO: thread safety */
        char *old_kite_dylib_path = kite_dylib_path;
#ifndef HAVE_GC_H
        kite_dylib_path = malloc(strlen(old_kite_dylib_path) + strlen($<stringValue>3) + 2);
#else
        kite_dylib_path = GC_malloc(strlen(old_kite_dylib_path) + strlen($<stringValue>3) + 2);
#endif
        strcpy(kite_dylib_path, old_kite_dylib_path);
        strcat(kite_dylib_path, ":");
        strcat(kite_dylib_path, $<stringValue>3);
    }
    else
    {
        /* TODO: throw exception if no valid option found. */
    }
};

import_statement:    IMPORT STRING_VALUE     {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = kite_top_stack(compiler->obj_created);
        kite_opcode_t *opc;
        
#ifndef HAVE_GC_H
        kite_funccall_info_t *fcall = (kite_funccall_info_t*)malloc(sizeof(kite_funccall_info_t));
#else
        kite_funccall_info_t *fcall = (kite_funccall_info_t*)GC_malloc(sizeof(kite_funccall_info_t));
#endif /* HAVE_GC_H */        
        fcall->file = compiler->file;
        fcall->line = @2.first_line;
        fcall->locals = NULL;
        fcall->last_deref = NULL;
        fcall->last_deref_obj = NULL;
        fcall->sys_or_user = 0;
        fcall->this = obj;
        kite_push_stack(&thd->func_stack, FALSE, fcall);
        
        /* Skip over import statement when executing so that the code inside
           the import doesn't get executed twice. */
        COMPILE_INSTRUCTION2(thd, compiler, kite_compile_jump_uncond(NULL), @2.first_line, &obj->builtin_data.funcvalue.funcptr);
        opc = (void*)thd->vm->cur_text_segment - sizeof(kite_opcode_t);
        
        kite_dereference_and_load(thd, $<stringValue>2);
        
        COMPILE_INSTRUCTION2(thd, compiler, kite_compile_nop(), @2.first_line, &obj->builtin_data.funcvalue.funcptr);
        opc = (void*)opc + (long)thd->vm->text_segment;
        opc->op_arg.jumpto = (void*)thd->vm->cur_text_segment - sizeof(kite_opcode_t);
        
        free($<stringValue>2);

        if (thd->exception) return 0;
        
        kite_pop_stack(&thd->func_stack);
#ifndef HAVE_GC_H
        free(fcall);
#endif /* HAVE_GC_H */
    }
    ;

break_statement:    BREAK
    {
        (*kite_compiler_actions[COMPILE_BREAK])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    ;

continue_statement: CONTINUE
    {
        (*kite_compiler_actions[COMPILE_CONTINUE])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    ;

return_statement: RETURN
    {
        (*kite_compiler_actions[COMPILE_RETURN])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    ;
        
optional_docstring:
    {
        $<stringValue>$ = strdup("");
    }
    | DOCSTRING {
        $<stringValue>$ = $<stringValue>1;
    }
    ;
        
method_declaration: METHOD SYMBOL_NAME optional_param_name_list optional_docstring {
        (*kite_compiler_actions[COMPILE_CREATE_FUNC])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<internalValue>3);
    } '[' input ']' {
        (*kite_compiler_actions[COMPILE_CREATE_FUNC_BODY])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>2, $<stringValue>4);
        free($<stringValue>2);
        free($<stringValue>4);
    }
    ;

anon_method_declaration: METHOD optional_param_name_list     {
        (*kite_compiler_actions[COMPILE_CREATE_ANON_FUNC])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<internalValue>2);
    } '[' input ']' {
        (*kite_compiler_actions[COMPILE_CREATE_ANON_FUNC_BODY])((kite_compiler_t*)yyget_extra(parm), @4.first_line);
    }
    ;
        
optional_global:    {
        $<intValue>$ = 0;
    }
    | GLOBAL    {
        $<intValue>$ = 1;
    }
    ;
    
optional_param_name_list:    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        $<internalValue>$ = kite_new_list(thd);
    }
    | param_name_list    {
        $<internalValue>$ = $<internalValue>1;
    }
    ;
    
param_name_list:    '(' ')'    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        $<internalValue>$ = kite_new_list(thd);
    }
    | '(' param_elements ')'    {
        $<internalValue>$ = $<internalValue>2;
    }
    ;
    
param_elements: SYMBOL_NAME optional_docstring   {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        checkReserved(parm, $<stringValue>1);
        $<internalValue>$ = kite_new_list(thd);
        $<internalValue>$->builtin_data.listvalue.car = 
            kite_new_ident(thd, $<stringValue>1);
        kite_set_docstring($<internalValue>$->builtin_data.listvalue.car, $<stringValue>2);
        free($<stringValue>1);
        free($<stringValue>2);
    }
    | SYMBOL_NAME optional_docstring ',' param_elements {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        checkReserved(parm, $<stringValue>1);
        $<internalValue>$ = kite_new_list(thd);
        $<internalValue>$->builtin_data.listvalue.car = 
            kite_new_ident(thd, $<stringValue>1);
        $<internalValue>$->builtin_data.listvalue.cdr = $<internalValue>4;
        kite_set_docstring($<internalValue>$->builtin_data.listvalue.car, $<stringValue>2);
        free($<stringValue>1);
        free($<stringValue>2);
    }
    ;
    
class_definition:    CLASS SYMBOL_NAME {
        (*kite_compiler_actions[COMPILE_CREATE_CLASS])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>2);
    } optional_from optional_docstring '[' class_body ']'    {
        (*kite_compiler_actions[COMPILE_CREATE_CLASS_BODY])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>2, $<stringValue>5);
        free($<stringValue>2);
        free($<stringValue>5);
    }
    ;
    
optional_from:        {
        (*kite_compiler_actions[COMPILE_FROM_DEFAULT])((kite_compiler_t*)yyget_extra(parm), 0);
    }
    | from_statement
    ;

from_statement:
    FROM variable_immediate    {
        (*kite_compiler_actions[COMPILE_FROM_GIVEN_CLASS])((kite_compiler_t*)yyget_extra(parm), 0, $<internalValue>2);
    }
    ;
    
variable_immediate: SYMBOL_NAME {
        $<internalValue>$ =
            (*kite_compiler_actions[COMPILE_IMMEDIATE_DEREF_1])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>1);
        free($<stringValue>1);
    }
    | variable_immediate '.' SYMBOL_NAME {
        $<internalValue>$ =
            (*kite_compiler_actions[COMPILE_IMMEDIATE_DEREF_2])((kite_compiler_t*)yyget_extra(parm), @3.first_line, $<internalValue>1, $<stringValue>3);
        free($<stringValue>3);
    }
    ;
    
class_body:    
    | class_elements
    ;

class_elements:    class_element
    | class_elements ',' class_element
    ;
    
class_element:    constructor_definition
    | destructor_definition
    | property_definition
    | operator_definition
    | class_definition
    | data_manipulation_statement
    ;

constructor_definition:    CONSTRUCT optional_param_name_list {
        (*kite_compiler_actions[COMPILE_CONSTRUCTOR])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<internalValue>2);
    } optional_docstring '[' input ']' {
        (*kite_compiler_actions[COMPILE_CONSTRUCTOR_BODY])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<internalValue>2, $<stringValue>4);
        free($<stringValue>4);
    }
    ;
    
destructor_definition: DESTRUCT {
        (*kite_compiler_actions[COMPILE_DESTRUCTOR])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    } optional_docstring '[' input ']' {
        (*kite_compiler_actions[COMPILE_DESTRUCTOR_BODY])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>3);
        free($<stringValue>3);
    }
    ;
    
property_definition:    optional_global PROPERTY SYMBOL_NAME optional_docstring   {
        (*kite_compiler_actions[COMPILE_PROPERTY_CREATE_WITH_DEFAULT_VAL])((kite_compiler_t*)yyget_extra(parm), @2.first_line, $<intValue>1, $<stringValue>3, $<stringValue>4);
        free($<stringValue>3);
        free($<stringValue>4);
    }
    | optional_global PROPERTY SYMBOL_NAME optional_docstring '=' {
            (*kite_compiler_actions[COMPILE_PROPERTY_CREATE_WITH_SPECIFIED_VAL])((kite_compiler_t*)yyget_extra(parm), @2.first_line, $<intValue>1, $<stringValue>3, $<stringValue>4);
            free($<stringValue>3);
            free($<stringValue>4);
        } data_manipulation_statement {
            (*kite_compiler_actions[COMPILE_PROPERTY_CREATE_WITH_SPECIFIED_VAL_RHS])((kite_compiler_t*)yyget_extra(parm), @5.first_line);
        }
    ;

operator_definition:    OPERATOR operator_type param_name_list {
    (*kite_compiler_actions[COMPILE_OPERATOR])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<internalValue>3);
} optional_docstring '[' input ']' {
    (*kite_compiler_actions[COMPILE_OPERATOR_BODY])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<intValue>2, $<stringValue>5);
    free($<stringValue>5);
}
    ;
    
operator_type:    PLUS    {
        $<intValue>$ = OP_ADD;
    }
    | MINUS    {
        $<intValue>$ = OP_SUBTRACT;
    }
    | MULTIPLY     {
        $<intValue>$ = OP_MULTIPLY;
    }
    | DIVIDE    {
        $<intValue>$ = OP_DIVIDE;
    }
    | MOD    {
        $<intValue>$ = OP_MODULUS;
    }
    | UNARY_PLUS    {
        $<intValue>$ = OP_UNARY_PLUS;
    }
    | UNARY_MINUS    {
        $<intValue>$ = OP_UNARY_MINUS;
    }
    | MAP    {
        $<intValue>$ = OP_MAP;
    }
    | REDUCE    {
        $<intValue>$ = OP_REDUCE;
    }
    | ARRAY_DEREF    {
        $<intValue>$ = OP_ARRAY_DEREF;
    }
    | ARRAY_SET {
        $<intValue>$ = OP_ARRAY_SET;
    }
    | EQUALS    {
        $<intValue>$ = OP_EQUALS;
    }
    | NOT_EQUALS    {
        $<intValue>$ = OP_NOT_EQUALS;
    }
    | LESS_THAN    {
        $<intValue>$ = OP_LESS_THAN;
    }
    | GREATER_THAN    {
        $<intValue>$ = OP_GREATER_THAN;
    }
    | LESS_OR_EQUALS    {
        $<intValue>$ = OP_LESS_OR_EQUALS;
    }
    | GREATER_OR_EQUALS    {
        $<intValue>$ = OP_GREATER_OR_EQUALS;
    }
    | AND    {
        $<intValue>$ = OP_AND;
    }
    | OR    {
        $<intValue>$ = OP_OR;
    }
    | NOT    {
        $<intValue>$ = OP_NOT;
    }
    | XOR    {
        $<intValue>$ = OP_XOR;
    }
    | LEFT_SHIFT    {
        $<intValue>$ = OP_LEFT_SHIFT;
    }
    | RIGHT_SHIFT    {
        $<intValue>$ = OP_RIGHT_SHIFT;
    }
    | METHOD_CALL    {
        $<intValue>$ = OP_METHOD_CALL;
    }
    | PROPERTY       {
        $<intValue>$ = OP_PROPERTY;
    }
    ;

data_manipulation_statement:    assignment_statement
    | method_declaration
    | eval_statement
    ;
    
assignment_statement:    deref_statement {
    $<opValue>$ = (*kite_compiler_actions[COMPILE_ASSIGNMENT_LHS])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
} '=' data_manipulation_statement {
        (*kite_compiler_actions[COMPILE_ASSIGNMENT_RHS])((kite_compiler_t*)yyget_extra(parm), @3.first_line, $<opValue>1, $<opValue>2);
    }
    | or_statement
    ;
    
or_statement:        xor_statement {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_OR_LHS])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    } OR or_statement {
        (*kite_compiler_actions[COMPILE_OR_RHS])((kite_compiler_t*)yyget_extra(parm), @3.first_line, $<opValue>2);
    }
    | xor_statement
    ;

xor_statement:    xor_statement XOR and_statement {
        (*kite_compiler_actions[COMPILE_XOR])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | and_statement
    ;
    
and_statement:    bitwise_statement {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_AND_LHS])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    } AND and_statement {
        (*kite_compiler_actions[COMPILE_AND_RHS])((kite_compiler_t*)yyget_extra(parm), @3.first_line, $<opValue>2);
    }
    | bitwise_statement
    ;

bitwise_statement:    bitwise_statement LEFT_SHIFT equality_statement {
        (*kite_compiler_actions[COMPILE_LEFT_SHIFT])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | bitwise_statement RIGHT_SHIFT equality_statement {
        (*kite_compiler_actions[COMPILE_RIGHT_SHIFT])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | equality_statement
    ;
    
equality_statement:    equality_statement EQUALS less_greater_statement {
        (*kite_compiler_actions[COMPILE_EQUALS])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | equality_statement NOT_EQUALS less_greater_statement {
        (*kite_compiler_actions[COMPILE_NOT_EQUALS])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | equality_statement IS less_greater_statement {
        (*kite_compiler_actions[COMPILE_IS])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | equality_statement ISOF less_greater_statement {
        (*kite_compiler_actions[COMPILE_ISOF])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | less_greater_statement
    ;
    
less_greater_statement:    less_greater_statement '<' add_sub_statement {
        (*kite_compiler_actions[COMPILE_LESS_THAN])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | less_greater_statement '>' add_sub_statement {
        (*kite_compiler_actions[COMPILE_GREATER_THAN])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | less_greater_statement LESS_OR_EQUALS add_sub_statement {
        (*kite_compiler_actions[COMPILE_LESS_EQUALS])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | less_greater_statement GREATER_OR_EQUALS add_sub_statement {
        (*kite_compiler_actions[COMPILE_GREATER_EQUALS])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | add_sub_statement
    ;
    
add_sub_statement:    add_sub_statement '+' multdiv_statement {
        (*kite_compiler_actions[COMPILE_ADD])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | add_sub_statement '-' multdiv_statement {
        (*kite_compiler_actions[COMPILE_SUBTRACT])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | multdiv_statement
    ;
    
multdiv_statement:    multdiv_statement '*' unary_statement {
        (*kite_compiler_actions[COMPILE_MULTIPLY])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | multdiv_statement '/' unary_statement {
        (*kite_compiler_actions[COMPILE_DIVIDE])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | multdiv_statement '%' unary_statement {
        (*kite_compiler_actions[COMPILE_MODULUS])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | unary_statement
    ;
    
unary_statement: '+' unary_statement {
        (*kite_compiler_actions[COMPILE_UNARY_PLUS])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    | '-' unary_statement {
        (*kite_compiler_actions[COMPILE_UNARY_MINUS])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    | NOT unary_statement {
        (*kite_compiler_actions[COMPILE_NOT])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    | map_reduce_statement
    ;
    
map_reduce_statement:    deref_statement MAP deref_statement {
        (*kite_compiler_actions[COMPILE_MAP])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | deref_statement REDUCE deref_statement {
        (*kite_compiler_actions[COMPILE_REDUCE])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    | SYMBOL_NAME { 
        (*kite_compiler_actions[COMPILE_FUNC_CALL_BEGIN])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>1);
        free($<stringValue>1);
    } param_list {
        (*kite_compiler_actions[COMPILE_FUNC_CALL_ARGS])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    | deref_statement
    ;
    
deref_statement:    deref_statement '|' SYMBOL_NAME {
        (*kite_compiler_actions[COMPILE_DEREF_AND_CALL_BEGIN])((kite_compiler_t*)yyget_extra(parm), @2.first_line, $<stringValue>3);
        free($<stringValue>3);
    } optional_param_list {
        (*kite_compiler_actions[COMPILE_DEREF_AND_CALL_ARGS])((kite_compiler_t*)yyget_extra(parm), @3.first_line, $<opValue>5);
        $<opValue>$ = $<opValue>1;
    }
    | deref_statement '[' data_manipulation_statement ']'    {
        (*kite_compiler_actions[COMPILE_ARRAY_DEREF])((kite_compiler_t*)yyget_extra(parm), @3.first_line);
        $<opValue>$ = $<opValue>1;
    }
    | deref_statement '.' SYMBOL_NAME {
        (*kite_compiler_actions[COMPILE_DEREF_2])((kite_compiler_t*)yyget_extra(parm), @3.first_line, $<stringValue>3);
        free($<stringValue>3);
        $<opValue>$ = $<opValue>1;
    }
    | SYMBOL_NAME {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_DEREF_1])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>1);
        free($<stringValue>1);
    }
    | constant_statement {
        $<opValue>$ = $<opValue>1;
    }
    ;

optional_param_list: {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_EMPTY_PARAM_LIST])((kite_compiler_t*)yyget_extra(parm), @0.first_line);
    }
    | param_list {
        $<opValue>$ = $<opValue>1;
    }
    ;

param_list:    '(' ')' {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_EMPTY_PARAM_LIST])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    | '(' params_list ')' {
        $<opValue>$ = $<opValue>2;
    }
    ;

params_list: data_manipulation_statement    {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_SINGLE_FUNC_PARAM])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    | data_manipulation_statement ',' params_list    {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_COMBINE_FUNC_PARAMS])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    ;
    
var_name:    SYMBOL_NAME {
        /*kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_opcode_t *op = kite_compile_push(kite_new_ident(thd, $<stringValue>1));
        COMPILE_INSTRUCTION(op, @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_deref_1(TRUE), @1.first_line);
        free($<stringValue>1);
        $<opValue>$ = op;*/
        $<opValue>$ = (*kite_compiler_actions[COMPILE_DEREF_1])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>1);
        free($<stringValue>1);
    }
    | var_name '.' SYMBOL_NAME {
        /*COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, $<stringValue>3)),
             @2.first_line);
        COMPILE_INSTRUCTION(kite_compile_deref_2(), @2.first_line);
        free($<stringValue>3);
        $<opValue>$ = $<opValue>1;*/
        (*kite_compiler_actions[COMPILE_DEREF_2])((kite_compiler_t*)yyget_extra(parm), @3.first_line, $<stringValue>3);
        free($<stringValue>3);
        $<opValue>$ = $<opValue>1;
    }
    ;
    
constant_statement:    STRING_VALUE     {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_CONSTANT_STRING])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>1);
        free($<stringValue>1);
    }
    | INT_VALUE {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_CONSTANT_INTEGER])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<intValue>1);
    }
    | FLOAT_VALUE {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_CONSTANT_FLOAT])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<doubleValue>1);
    }
    | BOOL_VALUE {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_CONSTANT_BOOL])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<intValue>1);
    }
    | NULL_VALUE {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_CONSTANT_NULL])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    | INT_VALUE ':' INT_VALUE
    {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_CONSTANT_RANGE])((kite_compiler_t*)yyget_extra(parm), @2.first_line, $<intValue>1, $<intValue>3, 1);
    }
    | INT_VALUE ':' INT_VALUE ':' INT_VALUE
    {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_CONSTANT_RANGE])((kite_compiler_t*)yyget_extra(parm), @2.first_line, $<intValue>1, $<intValue>3, $<intValue>5);
    }
    | REGEX_VALUE {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_CONSTANT_REGEX])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>1, FALSE);
        free($<stringValue>1);
    }
    | REGEX_VALUE_CI {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_CONSTANT_REGEX])((kite_compiler_t*)yyget_extra(parm), @1.first_line, $<stringValue>1, TRUE);
        free($<stringValue>1);
    }
    | list_value { $<opValue>$ = NULL; }
    | construct_object_statement { $<opValue>$ = NULL; }
    | anon_method_declaration { $<opValue>$ = NULL; }
    | '(' data_manipulation_statement ')' { $<opValue>$ = NULL; }
    ;

construct_object_statement:    MAKE var_name param_list    {
        (*kite_compiler_actions[COMPILE_MAKE])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    ;
    
list_value: '[' ']' {
        (*kite_compiler_actions[COMPILE_ZERO_ELEMENT_LIST])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    | '[' list_elements ']'
    ;
    
list_elements: list_element    {
        (*kite_compiler_actions[COMPILE_ONE_ELEMENT_LIST])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    | list_element ',' list_elements    {
        (*kite_compiler_actions[COMPILE_TWO_ELEMENT_LIST])((kite_compiler_t*)yyget_extra(parm), @2.first_line);
    }
    ;

list_element:    or_statement
    ;
    
decide_statement:    DECIDE {
        (*kite_compiler_actions[COMPILE_DECIDE_BEGIN])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    } '[' decide_contents ']' {
        (*kite_compiler_actions[COMPILE_DECIDE_END])((kite_compiler_t*)yyget_extra(parm), @5.first_line);
    }
    ;
    
decide_contents:    decide_condition
    | decide_contents ',' decide_condition
    ;
    
decide_condition:    constant_statement {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_DECIDE_CONDITION_BEGIN])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    } '[' input ']'    {
        (*kite_compiler_actions[COMPILE_DECIDE_CONDITION_END])((kite_compiler_t*)yyget_extra(parm), @5.first_line, $<opValue>2);
    }
    ;
    
eval_statement:    EVAL {
        (*kite_compiler_actions[COMPILE_EVAL_BEGIN])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    } or_statement    {
        (*kite_compiler_actions[COMPILE_EVAL_END])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    }
    ;

loop_statement:    loop_type '[' input ']'    {
        (*kite_compiler_actions[COMPILE_LOOP_WRAPPER])((kite_compiler_t*)yyget_extra(parm), @4.first_line, $<opValue>1);
    }
    ;

loop_type:    WHILE {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_WHILE_BEGIN])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    } constant_statement    {
        (*kite_compiler_actions[COMPILE_WHILE_END])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
        $<opValue>$ = $<opValue>2;
    }
    | UNTIL {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_UNTIL_BEGIN])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    } constant_statement    {
        (*kite_compiler_actions[COMPILE_UNTIL_END])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
        $<opValue>$ = $<opValue>2;
    }
    ;

run_catch_statement: RUN {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_RUN_BEGIN])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    } '[' input ']' {
        $<opValue>$ = (*kite_compiler_actions[COMPILE_RUN_END])((kite_compiler_t*)yyget_extra(parm), @1.first_line);
    } CATCH {
        (*kite_compiler_actions[COMPILE_CATCH_BEGIN])((kite_compiler_t*)yyget_extra(parm), @7.first_line, $<opValue>2);
    } '[' input ']' {
        (*kite_compiler_actions[COMPILE_CATCH_END])((kite_compiler_t*)yyget_extra(parm), @7.first_line, $<opValue>6);
    }
    ;
    
%%

void dirty_yyerror(void *yyparse_param, char *str) {
    kite_object_t *exc;
    kite_compiler_t *comp = (kite_compiler_t*)yyget_extra(yyparse_param);
    kite_thread_t *thd = comp->thd;
    
    kite_funccall_info_t *entry = calloc(sizeof(kite_funccall_info_t), 1);
    entry->this = NULL;
    entry->file = comp->file;
    entry->line = comp->currentLine;
    kite_push_stack(&thd->func_stack, FALSE, entry);

    exc = kite_new_exception(thd, "System.exceptions.SyntaxError", 
                             str);
    kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);    
}

void checkReserved(void *p, char *name) {
    if (!strcmp(name, "__exc") ||
        !strcmp(name, "this")) {
        dirty_yyerror(p, "Cannot use reserved names for parameters.");
    }
}
